@@ -54,6 +54,8 @@
 
 #define FTP_INTERFACE  "org.openobex.FileTransfer"
 
+#define OBEX_IO_ERROR obex_io_error_quark()
+
 static guint64 counter = 0;
 
 static unsigned char pcsuite_uuid[] = { 0x00, 0x00, 0x50, 0x05, 0x00, 0x00,
@@ -85,7 +87,16 @@ struct agent_data {
 	struct pending_data *pending;
 };
 
-static void session_prepare_put(struct session_data *session, void *data);
+static void session_prepare_put(struct session_data *session, GError *err,
+								void *data);
+static void session_terminate_transfer(struct session_data *session,
+					struct transfer_data *transfer,
+					GError *gerr);
+
+static GQuark obex_io_error_quark(void)
+{
+	return g_quark_from_static_string("obex-io-error-quark");
+}
 
 struct session_data *session_ref(struct session_data *session)
 {
@@ -173,8 +184,10 @@ static void session_free(struct session_data *session)
 	if (session->obex != NULL)
 		gw_obex_close(session->obex);
 
-	if (session->sock > 2)
-		close(session->sock);
+	if (session->io != NULL) {
+		g_io_channel_shutdown(session->io, TRUE, NULL);
+		g_io_channel_unref(session->io);
+	}
 
 	if (session->path)
 		session_unregistered(session);
@@ -215,25 +228,29 @@ static void rfcomm_callback(GIOChannel *io, GError *err, gpointer user_data)
 		goto done;
 	}
 
+	/* do not close when gw_obex is using the fd */
+	g_io_channel_set_close_on_unref(session->io, FALSE);
+	g_io_channel_unref(session->io);
+	session->io = NULL;
+
 	fd = g_io_channel_unix_get_fd(io);
 
 	obex = gw_obex_setup_fd(fd, session->target,
 			session->target_len, NULL, NULL);
 
-	callback->session->sock = fd;
-	callback->session->obex = obex;
+	session->obex = obex;
 
 done:
-	callback->func(callback->session, callback->data);
+	callback->func(callback->session, err, callback->data);
 
 	session_unref(callback->session);
 
 	g_free(callback);
 }
 
-static int rfcomm_connect(const bdaddr_t *src,
-				const bdaddr_t *dst, uint8_t channel,
-					BtIOConnect function, gpointer user_data)
+static GIOChannel *rfcomm_connect(const bdaddr_t *src, const bdaddr_t *dst,
+					uint8_t channel, BtIOConnect function,
+					gpointer user_data)
 {
 	GIOChannel *io;
 	GError *err = NULL;
@@ -245,11 +262,11 @@ static int rfcomm_connect(const bdaddr_t *src,
 				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
 				BT_IO_OPT_INVALID);
 	if (io != NULL)
-		return 0;
+		return io;
 
 	error("%s", err->message);
 	g_error_free(err);
-	return -EIO;
+	return NULL;
 }
 
 static void search_callback(uint8_t type, uint16_t status,
@@ -259,6 +276,7 @@ static void search_callback(uint8_t type, uint16_t status,
 	unsigned int scanned, bytesleft = size;
 	int seqlen = 0;
 	uint8_t dataType, channel = 0;
+	GError *gerr = NULL;
 
 	if (status || type != SDP_SVC_SEARCH_ATTR_RSP)
 		goto failed;
@@ -309,8 +327,11 @@ static void search_callback(uint8_t type, uint16_t status,
 
 	callback->session->channel = channel;
 
-	if (rfcomm_connect(&callback->session->src, &callback->session->dst,
-					channel, rfcomm_callback, callback) == 0) {
+	callback->session->io = rfcomm_connect(&callback->session->src,
+						&callback->session->dst,
+						channel, rfcomm_callback,
+						callback);
+	if (callback->session->io != NULL) {
 		sdp_close(callback->sdp);
 		return;
 	}
@@ -318,7 +339,11 @@ static void search_callback(uint8_t type, uint16_t status,
 failed:
 	sdp_close(callback->sdp);
 
-	callback->func(callback->session, callback->data);
+	g_set_error(&gerr, OBEX_IO_ERROR, -EIO,
+					"Unable to find service record");
+	callback->func(callback->session, gerr, callback->data);
+	g_clear_error(&gerr);
+
 	session_unref(callback->session);
 	g_free(callback);
 }
@@ -343,6 +368,7 @@ static gboolean service_callback(GIOChannel *io, GIOCondition cond,
 	struct callback_data *callback = user_data;
 	sdp_list_t *search, *attrid;
 	uint32_t range = 0x0000ffff;
+	GError *gerr = NULL;
 
 	if (cond & (G_IO_NVAL | G_IO_ERR))
 		goto failed;
@@ -371,7 +397,11 @@ static gboolean service_callback(GIOChannel *io, GIOCondition cond,
 failed:
 	sdp_close(callback->sdp);
 
-	callback->func(callback->session, callback->data);
+	g_set_error(&gerr, OBEX_IO_ERROR, -EIO,
+					"Unable to find service record");
+	callback->func(callback->session, gerr, callback->data);
+	g_clear_error(&gerr);
+
 	session_unref(callback->session);
 	g_free(callback);
 	return FALSE;
@@ -418,7 +448,6 @@ struct session_data *session_create(const char *source,
 		return NULL;
 
 	session->refcount = 1;
-	session->sock = -1;
 	session->channel = channel;
 
 	session->conn = dbus_bus_get(DBUS_BUS_SESSION, NULL);
@@ -465,8 +494,11 @@ struct session_data *session_create(const char *source,
 	callback->data = user_data;
 
 	if (session->channel > 0) {
-		err = rfcomm_connect(&session->src, &session->dst,
-				session->channel, rfcomm_callback, callback);
+		session->io = rfcomm_connect(&session->src, &session->dst,
+							session->channel,
+							rfcomm_callback,
+							callback);
+		err = (session->io == NULL) ? -EINVAL : 0;
 	} else {
 		callback->sdp = service_connect(&session->src, &session->dst,
 						service_callback, callback);
@@ -717,7 +749,7 @@ static const GMarkupParser parser = {
 };
 
 static void list_folder_callback(struct session_data *session,
-					void *user_data)
+					GError *err, void *user_data)
 {
 	struct transfer_data *transfer = session->pending->data;
 	GMarkupParseContext *ctxt;
@@ -757,7 +789,8 @@ done:
 	session->msg = NULL;
 }
 
-static void get_file_callback(struct session_data *session, void *user_data)
+static void get_file_callback(struct session_data *session, GError *err,
+							void *user_data)
 {
 
 }
@@ -773,11 +806,18 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 
 	dbus_error_init(&derr);
 	if (dbus_set_error_from_message(&derr, reply)) {
+		GError *gerr = NULL;
+
 		error("Replied with an error: %s, %s",
 				derr.name, derr.message);
 		dbus_error_free(&derr);
 		dbus_message_unref(reply);
-		transfer_unregister(pending->transfer);
+
+		g_set_error(&gerr, OBEX_IO_ERROR, -ECANCELED, "%s",
+								derr.message);
+		session_terminate_transfer(session, pending->transfer, gerr);
+		g_clear_error(&gerr);
+
 		return;
 	}
 
@@ -792,10 +832,11 @@ static void session_request_reply(DBusPendingCall *call, gpointer user_data)
 		pending->transfer->name = g_strdup(name);
 	}
 
-	pending->cb(session, pending->transfer);
+	agent->pending = NULL;
+
+	pending->cb(session, NULL, pending->transfer);
 	dbus_message_unref(reply);
 	free_pending(pending);
-	agent->pending = NULL;
 
 	return;
 }
@@ -805,7 +846,7 @@ static gboolean session_request_proceed(gpointer data)
 	struct pending_data *pending = data;
 	struct transfer_data *transfer = pending->transfer;
 
-	pending->cb(transfer->session, transfer);
+	pending->cb(transfer->session, NULL, transfer);
 	free_pending(pending);
 
 	return FALSE;
@@ -853,12 +894,13 @@ static int session_request(struct session_data *session, session_callback_t cb,
 }
 
 static void session_terminate_transfer(struct session_data *session,
-					struct transfer_data *transfer)
+					struct transfer_data *transfer,
+					GError *gerr)
 {
 	struct session_callback *callback = session->callback;
 
 	if (callback) {
-		callback->func(session, callback->data);
+		callback->func(session, gerr, callback->data);
 		return;
 	}
 
@@ -899,12 +941,12 @@ done:
 
 	DBG("Transfer(%p) complete", transfer);
 
-	session_terminate_transfer(session, transfer);
+	session_terminate_transfer(session, transfer, NULL);
 }
 
 static void session_notify_error(struct session_data *session,
 				struct transfer_data *transfer,
-				const char *err)
+				GError *err)
 {
 	struct agent_data *agent = session->agent;
 	DBusMessage *message;
@@ -914,7 +956,7 @@ static void session_notify_error(struct session_data *session,
 
 		reply = g_dbus_create_error(session->msg,
 					"org.openobex.Error.Failed",
-					err);
+					"%s", err->message);
 		g_dbus_send_message(session->conn, reply);
 
 		dbus_message_unref(session->msg);
@@ -933,15 +975,15 @@ static void session_notify_error(struct session_data *session,
 
 	dbus_message_append_args(message,
 			DBUS_TYPE_OBJECT_PATH, &transfer->path,
-			DBUS_TYPE_STRING, &err,
+			DBUS_TYPE_STRING, &err->message,
 			DBUS_TYPE_INVALID);
 
 	g_dbus_send_message(session->conn, message);
 
 done:
-	error("Transfer(%p) Error: %s", transfer, err);
+	error("Transfer(%p) Error: %s", transfer, err->message);
 
-	session_terminate_transfer(session, transfer);
+	session_terminate_transfer(session, transfer, err);
 }
 
 static void session_notify_progress(struct session_data *session,
@@ -991,6 +1033,7 @@ static void transfer_progress(struct transfer_data *transfer, gint64 transferred
 				int err, void *user_data)
 {
 	struct session_data *session = user_data;
+	GError *gerr = NULL;
 
 	if (err != 0)
 		goto fail;
@@ -1000,18 +1043,25 @@ static void transfer_progress(struct transfer_data *transfer, gint64 transferred
 	return;
 
 fail:
-	session_notify_error(session, transfer,
+	g_set_error(&gerr, OBEX_IO_ERROR, err, "%s",
 			err > 0 ? OBEX_ResponseToString(err) : strerror(-err));
+	session_notify_error(session, transfer, gerr);
+	g_clear_error(&gerr);
 }
 
-static void session_prepare_get(struct session_data *session, void *data)
+static void session_prepare_get(struct session_data *session,
+				GError *err, void *data)
 {
 	struct transfer_data *transfer = data;
-	int err;
+	int ret;
 
-	err = transfer_get(transfer, transfer_progress, session);
-	if (err < 0) {
-		session_notify_error(session, transfer, strerror(-err));
+	ret = transfer_get(transfer, transfer_progress, session);
+	if (ret < 0) {
+		GError *gerr = NULL;
+
+		g_set_error(&gerr, OBEX_IO_ERROR, ret, "%s", strerror(-ret));
+		session_notify_error(session, transfer, gerr);
+		g_clear_error(&gerr);
 		return;
 	}
 
@@ -1077,7 +1127,7 @@ static DBusMessage *change_folder(DBusConnection *connection,
 	if (gw_obex_chdir(session->obex, folder, &err) == FALSE) {
 		return g_dbus_create_error(message,
 				"org.openobex.Error.Failed",
-				OBEX_ResponseToString(err));
+				"%s", OBEX_ResponseToString(err));
 	}
 
 	return dbus_message_new_method_return(message);
@@ -1099,7 +1149,7 @@ static DBusMessage *create_folder(DBusConnection *connection,
 	if (gw_obex_mkdir(session->obex, folder, &err) == FALSE)
 		return g_dbus_create_error(message,
 				"org.openobex.Error.Failed",
-				OBEX_ResponseToString(err));
+				"%s", OBEX_ResponseToString(err));
 
 	return dbus_message_new_method_return(message);
 }
@@ -1204,7 +1254,7 @@ static DBusMessage *delete(DBusConnection *connection,
 	if (gw_obex_delete(session->obex, file, &err) == FALSE) {
 		return g_dbus_create_error(message,
 				"org.openobex.Error.Failed",
-				OBEX_ResponseToString(err));
+				"%s", OBEX_ResponseToString(err));
 	}
 
 	return dbus_message_new_method_return(message);
@@ -1235,10 +1285,8 @@ int session_send(struct session_data *session, const char *filename,
 
 	transfer = transfer_register(session, filename, targetname, NULL,
 					NULL);
-	if (transfer == NULL) {
-		err = -EINVAL;
-		goto fail;
-	}
+	if (transfer == NULL)
+		return -EINVAL;
 
 	/* Transfer should start if it is the first in the pending list */
 	if (transfer != session->pending->data)
@@ -1335,14 +1383,20 @@ void session_set_data(struct session_data *session, void *priv)
 	session->priv = priv;
 }
 
-static void session_prepare_put(struct session_data *session, void *data)
+static void session_prepare_put(struct session_data *session,
+				GError *err, void *data)
 {
 	struct transfer_data *transfer = data;
-	int err;
+	int ret;
 
-	err = transfer_put(transfer, transfer_progress, session);
-	if (err < 0) {
-		session_notify_error(session, transfer, strerror(-err));
+	ret = transfer_put(transfer, transfer_progress, session);
+	if (ret < 0) {
+		GError *gerr = NULL;
+
+		g_set_error(&gerr, OBEX_IO_ERROR, ret, "%s (%d)",
+							strerror(-ret), -ret);
+		session_notify_error(session, transfer, gerr);
+		g_clear_error(&gerr);
 		return;
 	}
 
