@@ -315,7 +315,7 @@ static void vcard_printf_slash_tag(GString *vcards, uint8_t format,
 
 static void vcard_printf_email(GString *vcards, uint8_t format,
 					const char *address,
-					enum phonebook_email_type category)
+					enum phonebook_field_type category)
 {
 	const char *category_string = "";
 	char field[LEN_MAX];
@@ -326,13 +326,13 @@ static void vcard_printf_email(GString *vcards, uint8_t format,
 		return;
 	}
 	switch (category){
-	case EMAIL_TYPE_HOME:
+	case FIELD_TYPE_HOME:
 		if (format == FORMAT_VCARD21)
 			category_string = "INTERNET;HOME";
 		else if (format == FORMAT_VCARD30)
 			category_string = "TYPE=INTERNET;TYPE=HOME";
 		break;
-	case EMAIL_TYPE_WORK:
+	case FIELD_TYPE_WORK:
 		if (format == FORMAT_VCARD21)
 			category_string = "INTERNET;WORK";
 		else if (format == FORMAT_VCARD30)
@@ -349,6 +349,41 @@ static void vcard_printf_email(GString *vcards, uint8_t format,
 	vcard_printf(vcards,"EMAIL;%s:%s", category_string, field);
 }
 
+static void vcard_printf_url(GString *vcards, uint8_t format,
+					const char *url,
+					enum phonebook_field_type category)
+{
+	const char *category_string = "";
+
+	if (!url || strlen(url) == 0) {
+		vcard_printf(vcards, "URL:");
+		return;
+	}
+
+	switch (category) {
+	case FIELD_TYPE_HOME:
+		if (format == FORMAT_VCARD21)
+			category_string = "INTERNET;HOME";
+		else if (format == FORMAT_VCARD30)
+			category_string = "TYPE=INTERNET;TYPE=HOME";
+		break;
+	case FIELD_TYPE_WORK:
+		if (format == FORMAT_VCARD21)
+			category_string = "INTERNET;WORK";
+		else if (format == FORMAT_VCARD30)
+			category_string = "TYPE=INTERNET;TYPE=WORK";
+		break;
+	default:
+		if (format == FORMAT_VCARD21)
+			category_string = "INTERNET";
+		else if (format == FORMAT_VCARD30)
+			category_string = "TYPE=INTERNET";
+		break;
+	}
+
+	vcard_printf(vcards,"URL;%s:%s", category_string, url);
+}
+
 static gboolean org_fields_present(struct phonebook_contact *contact)
 {
 	if (contact->company && strlen(contact->company))
@@ -374,7 +409,7 @@ static void vcard_printf_org(GString *vcards,
 
 static void vcard_printf_address(GString *vcards, uint8_t format,
 					const char *address,
-					enum phonebook_address_type category)
+					enum phonebook_field_type category)
 {
 	char buf[LEN_MAX];
 	char field[ADDR_FIELD_AMOUNT][LEN_MAX];
@@ -388,13 +423,13 @@ static void vcard_printf_address(GString *vcards, uint8_t format,
 	}
 
 	switch (category) {
-	case ADDR_TYPE_HOME:
+	case FIELD_TYPE_HOME:
 		if (format == FORMAT_VCARD21)
 			category_string = "HOME";
 		else if (format == FORMAT_VCARD30)
 			category_string = "TYPE=HOME";
 		break;
-	case ADDR_TYPE_WORK:
+	case FIELD_TYPE_WORK:
 		if (format == FORMAT_VCARD21)
 			category_string = "WORK";
 		else if (format == FORMAT_VCARD30)
@@ -457,10 +492,16 @@ static void vcard_printf_end(GString *vcards)
 void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 					uint64_t filter, uint8_t format)
 {
-	if (format == FORMAT_VCARD30)
+	if (format == FORMAT_VCARD30 && filter)
 		filter |= (FILTER_VERSION | FILTER_FN | FILTER_N | FILTER_TEL);
-	else if (format == FORMAT_VCARD21)
+	else if (format == FORMAT_VCARD21 && filter)
 		filter |= (FILTER_VERSION | FILTER_N | FILTER_TEL);
+	else
+		filter = (FILTER_VERSION | FILTER_UID | FILTER_N | FILTER_FN |
+				FILTER_TEL | FILTER_EMAIL | FILTER_ADR |
+				FILTER_BDAY | FILTER_NICKNAME | FILTER_URL |
+				FILTER_PHOTO | FILTER_ORG | FILTER_ROLE |
+				FILTER_TITLE | FILTER_X_IRMC_CALL_DATETIME);
 
 	vcard_printf_begin(vcards, format);
 
@@ -481,9 +522,9 @@ void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 							TEL_TYPE_OTHER);
 
 		for (; l; l = l->next) {
-			struct phonebook_number *number = l->data;
+			struct phonebook_field *number = l->data;
 
-			vcard_printf_number(vcards, format, number->tel, 1,
+			vcard_printf_number(vcards, format, number->text, 1,
 								number->type);
 		}
 	}
@@ -493,11 +534,11 @@ void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 
 		if (g_slist_length(l) == 0)
 			vcard_printf_email(vcards, format, NULL,
-							EMAIL_TYPE_OTHER);
+							FIELD_TYPE_OTHER);
 
 		for (; l; l = l->next){
-			struct phonebook_email *email = l->data;
-			vcard_printf_email(vcards, format, email->address,
+			struct phonebook_field *email = l->data;
+			vcard_printf_email(vcards, format, email->text,
 								email->type);
 		}
 	}
@@ -507,11 +548,11 @@ void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 
 		if (g_slist_length(l) == 0)
 			vcard_printf_address(vcards, format, NULL,
-							ADDR_TYPE_OTHER);
+							FIELD_TYPE_OTHER);
 
 		for (; l; l = l->next) {
-			struct phonebook_address *addr = l->data;
-			vcard_printf_address(vcards, format, addr->addr,
+			struct phonebook_field *addr = l->data;
+			vcard_printf_address(vcards, format, addr->text,
 								addr->type);
 		}
 	}
@@ -524,9 +565,18 @@ void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 		vcard_printf_slash_tag(vcards, format, "NICKNAME", NULL,
 							contact->nickname);
 
-	if (filter & FILTER_URL)
-		vcard_printf_slash_tag(vcards, format, "URL", "INTERNET",
-							contact->website);
+	if (filter & FILTER_URL) {
+		GSList *l = contact->urls;
+
+		if (g_slist_length(l) == 0)
+			vcard_printf_url(vcards, format, NULL,
+							FIELD_TYPE_OTHER);
+
+		for (; l; l = l->next) {
+			struct phonebook_field *url = l->data;
+			vcard_printf_url(vcards, format, url->text, url->type);
+		}
+	}
 
 	if (filter & FILTER_PHOTO)
 		vcard_printf_tag(vcards, format, "PHOTO", NULL,
@@ -547,28 +597,13 @@ void phonebook_add_contact(GString *vcards, struct phonebook_contact *contact,
 	vcard_printf_end(vcards);
 }
 
-static void number_free(gpointer data, gpointer user_data)
+
+static void field_free(gpointer data, gpointer user_data)
 {
-	struct phonebook_number *number = data;
+	struct phonebook_field *field = data;
 
-	g_free(number->tel);
-	g_free(number);
-}
-
-static void email_free(gpointer data, gpointer user_data)
-{
-	struct phonebook_email *email = data;
-
-	g_free(email->address);
-	g_free(email);
-}
-
-static void address_free(gpointer data, gpointer user_data)
-{
-	struct phonebook_address *addr = data;
-
-	g_free(addr->addr);
-	g_free(addr);
+	g_free(field->text);
+	g_free(field);
 }
 
 void phonebook_contact_free(struct phonebook_contact *contact)
@@ -576,15 +611,18 @@ void phonebook_contact_free(struct phonebook_contact *contact)
 	if (contact == NULL)
 		return;
 
-	g_slist_foreach(contact->numbers, number_free, NULL);
+	g_slist_foreach(contact->numbers, field_free, NULL);
 	g_slist_free(contact->numbers);
 
-	g_slist_foreach(contact->emails, email_free, NULL);
+	g_slist_foreach(contact->emails, field_free, NULL);
 	g_slist_free(contact->emails);
 
-	g_slist_foreach(contact->addresses, address_free, NULL);
+	g_slist_foreach(contact->addresses, field_free, NULL);
 	g_slist_free(contact->addresses);
 
+	g_slist_foreach(contact->urls, field_free, NULL);
+	g_slist_free(contact->urls);
+
 	g_free(contact->uid);
 	g_free(contact->fullname);
 	g_free(contact->given);
@@ -594,7 +632,6 @@ void phonebook_contact_free(struct phonebook_contact *contact)
 	g_free(contact->suffix);
 	g_free(contact->birthday);
 	g_free(contact->nickname);
-	g_free(contact->website);
 	g_free(contact->photo);
 	g_free(contact->company);
 	g_free(contact->department);
