@@ -317,8 +317,6 @@ static void obex_session_free(struct obex_session *os)
 {
 	sessions = g_slist_remove(sessions, os);
 
-	os_reset_session(os);
-
 	if (os->io)
 		g_io_channel_unref(os->io);
 
@@ -590,7 +588,8 @@ static int obex_read_stream(struct obex_session *os, obex_t *obex,
 
 	/* only write if both object and driver are valid */
 	if (os->object == NULL || os->driver == NULL) {
-		DBG("Stored %u bytes into temporary buffer", os->pending);
+		DBG("Stored %" PRIu64 " bytes into temporary buffer",
+								os->pending);
 		return 0;
 	}
 
@@ -639,10 +638,13 @@ static int obex_write_stream(struct obex_session *os,
 
 		len = MIN(os->size - os->offset, os->tx_mtu);
 		ptr = os->buf + os->offset;
+		flags = 0;
 		goto add_header;
 	}
 
-	len = os->driver->read(os->object, os->buf, os->tx_mtu, &hi);
+	ptr = os->buf + os->write_offset;
+	len = os->driver->read(os->object, ptr, os->tx_mtu - os->write_offset,
+								&hi, &flags);
 	if (len < 0) {
 		error("read(): %s (%zd)", strerror(-len), -len);
 		if (len == -EAGAIN)
@@ -655,18 +657,15 @@ static int obex_write_stream(struct obex_session *os,
 		return len;
 	}
 
-	ptr = os->buf;
-
 add_header:
 
 	hd.bs = ptr;
 
 	switch (hi) {
 	case OBEX_HDR_BODY:
-		flags = len ? OBEX_FL_STREAM_DATA : OBEX_FL_STREAM_DATAEND;
+		flags |= len ? OBEX_FL_STREAM_DATA : OBEX_FL_STREAM_DATAEND;
 		break;
 	case OBEX_HDR_APPARAM:
-		flags =  0;
 		break;
 	default:
 		error("read(): unkown header type %u", hi);
@@ -680,6 +679,11 @@ add_header:
 		os->buf = NULL;
 	}
 
+	if (flags & OBEX_FL_FIT_ONE_PACKET)
+		os->write_offset += len;
+	else
+		os->write_offset = 0;
+
 	os->offset += len;
 
 	return 0;
@@ -702,9 +706,10 @@ static gboolean handle_async_io(void *object, int flags, int err,
 		ret = obex_read_stream(os, os->obex, os->obj);
 
 proceed:
-	if (ret < 0)
+	if (ret < 0) {
+		os_set_response(os->obj, err);
 		OBEX_CancelRequest(os->obex, TRUE);
-	else
+	} else
 		OBEX_ResumeRequest(os->obex);
 
 	return FALSE;
@@ -795,7 +800,7 @@ static void cmd_get(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 	if (err < 0)
 		goto done;
 
-	if (os->size != OBJECT_SIZE_UNKNOWN) {
+	if (os->size != OBJECT_SIZE_UNKNOWN && os->size < UINT32_MAX) {
 		hd.bq4 = os->size;
 		OBEX_ObjectAddHeader(obex, obj,
 				OBEX_HDR_LENGTH, hd, 4, 0);
@@ -1004,7 +1009,7 @@ static gboolean check_put(obex_t *obex, obex_object_t *obj)
 
 		case OBEX_HDR_LENGTH:
 			os->size = hd.bq4;
-			DBG("OBEX_HDR_LENGTH: %d", os->size);
+			DBG("OBEX_HDR_LENGTH: %" PRIu64, os->size);
 			break;
 		case OBEX_HDR_TIME:
 			os->time = parse_iso8610((const char *) hd.bs, hlen);
@@ -1088,7 +1093,7 @@ static void cmd_put(struct obex_session *os, obex_t *obex, obex_object_t *obj)
 		return;
 	}
 
-	err = os->service->put(os, os->service_data);
+	err = os->service->put(os, obj, os->service_data);
 	if (err < 0)
 		os_set_response(obj, err);
 }
@@ -1224,6 +1229,8 @@ static void obex_handle_destroy(void *user_data)
 
 	os = OBEX_GetUserData(obex);
 
+	os_reset_session(os);
+
 	if (os->service && os->service->disconnect)
 		os->service->disconnect(os, os->service_data);
 
